package MyIR::Index;
use strict;
use warnings;
use base qw/Class::Accessor::Lvalue::Fast/;

use Storable ();

use MyIR::Document;
use MyIR::Util qw/ngram byte_lc intersect/;

__PACKAGE__->mk_accessors(qw/index/);

sub new {
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->index = {};
    return $self;
}

sub add_document {
    my ($self, $doc) = @_;

    my @words = ngram(
        byte_lc(join("\n", $doc->title, $doc->url, $doc->content))
    );

    my %seen;
    for my $w (@words) {
        if (not $seen{$w}++) {
            $self->add($w => $doc->doc_id);
        }
    }
}

sub add {
    my ($self, $w, $doc_id) = @_;
    if ($self->index->{$w}) {
        push @{$self->index->{$w}}, $doc_id;
    } else {
        $self->index->{$w} = [ $doc_id ]
    }
    return $doc_id;
}

sub retrieve_documents {
    my $self = shift;
    my @doc_ids = $self->retrieve(@_) or return;
    my @documents = map { MyIR::Document->new_from_id($_) } @doc_ids;
    return sort_documents(\@documents, @_)
}

sub retrieve {
    my ($self, $q) = @_;

    my @postings;
    for my $qt (split /\s+/, $q) {
        for my $w (ngram(byte_lc($qt))) {
            if (my $p = $self->get_postings($w)) {
                push @postings, $p;
            } else {
                return; # not found
            }
        }
    }

    @postings = sort { @$a <=> @$b } @postings ;

    my $p = shift @postings;
    while (@postings > 0 and defined $p) {
        $p = intersect($p, shift @postings);
    }

    @$p > 0 ? return @$p : return;
}

sub get_postings {
    my ($self, $w) = @_;
    return $self->index->{$w};
}


sub sort_documents {
    my $docs = shift;
    return sort { $b->doc_id <=> $a->doc_id } @$docs;
}

## FIXME: tf, タイトルに含まれるかどうか, ドメインでボーナス, 正規化
# sub sort_documents {
#     my ($docs, $q) = @_;

#     for my $d (@$docs) {
#         $d->score = 0;

#         for my $qt (byte_lc( split /\s+/, $q )) {
#             my $text = byte_lc(join "\n", $d->title, $d->url, $d->content);
#             my @tmp = ($text =~ m/$qt/g);
#             $d->score += @tmp;
#         }
#         # warn $d->score;
#     }

#     return sort { $b->score <=> $a->score } @$docs;
# }

sub save {
    my ($self, $file) = @_;
    Storable::store($self, $file) or die "Can't store index: $!";
}

sub load {
    my ($class, $file) = @_;
    return Storable::retrieve($file) or die "Can't retrieve index: $!";
}

sub dump {
    my $self = shift;
    require Data::Dumper;
    warn Data::Dumper::Dumper( $self->index );
}

1;
